@@ -18,7 +18,7 @@
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  *
- *  $Id: ephy-tab.c,v 1.168 2005/11/18 18:37:36 chpe Exp $
+ *  $Id: ephy-tab.c,v 1.172.2.2 2006/04/28 10:56:51 wbolster Exp $
  */
 
 #include "config.h"
@@ -58,6 +58,7 @@
 #include <gtk/gtkmenu.h>
 #include <gtk/gtkuimanager.h>
 #include <gtk/gtkclipboard.h>
+#include <libgnomevfs/gnome-vfs.h>
 #include <libgnomevfs/gnome-vfs-result.h>
 #include <libgnomevfs/gnome-vfs-monitor.h>
 #include <libgnomevfs/gnome-vfs-ops.h>
@@ -77,6 +78,8 @@
 
 struct _EphyTabPrivate
 {
+	guint id;
+
 	char *status_message;
 	char *link_message;
 	char *address;
@@ -145,7 +148,15 @@ typedef struct
 	char *features;
 } PopupInfo;
 
-static GObjectClass *parent_class = NULL;
+static GObjectClass *parent_class;
+
+/* We need to assign unique IDs to tabs, otherwise accels get confused in the
+ * tabs menu (bug #339548). We could use a serial #, but the ID is used in the
+ * action name which is stored in a GQuark and so we should use them sparingly.
+ */
+
+static GArray *tabs_id_array = NULL;
+static guint n_tabs = 0;
 
 /* internal functions, accessible only from this file */
 static void	ephy_tab_set_load_status	(EphyTab *tab,
@@ -535,7 +546,7 @@ popups_manager_add (EphyTab *tab,
 
 	popup = g_new0 (PopupInfo, 1);
 
-	popup->url = g_strdup (url);
+	popup->url = (url == NULL) ? NULL : g_strdup (url);
 	popup->name = g_strdup (name);
 	popup->features = g_strdup (features);
 
@@ -619,7 +630,7 @@ ephy_tab_get_popups_allowed (EphyTab *tab)
 	location = ephy_embed_get_location (embed, TRUE);
 	if (location == NULL) return FALSE; /* FALSE, TRUE... same thing */
 
-	response = ephy_permission_manager_test
+	response = ephy_permission_manager_test_permission
 		(permission_manager, location, EPT_POPUP);
 
 	switch (response)
@@ -651,14 +662,17 @@ popups_manager_show (PopupInfo *popup,
 	EphyEmbed *embed;
 	EphyEmbedSingle *single;
 
-	embed = ephy_tab_get_embed (tab);
+	/* Only show popup with non NULL url */
+	if (popup->url != NULL)
+	{
+		embed = ephy_tab_get_embed (tab);
 
-	single = EPHY_EMBED_SINGLE
-		(ephy_embed_shell_get_embed_single (embed_shell));
-
-	ephy_embed_single_open_window (single, embed, popup->url,
-				       popup->name, popup->features);
+		single = EPHY_EMBED_SINGLE
+			(ephy_embed_shell_get_embed_single (embed_shell));
 
+		ephy_embed_single_open_window (single, embed, popup->url,
+				popup->name, popup->features);
+	}
 	popups_manager_free_info (popup);
 }
 
@@ -757,7 +771,7 @@ ephy_tab_set_popups_allowed (EphyTab *tab,
 	permission = allowed ? EPHY_PERMISSION_ALLOWED
 			     : EPHY_PERMISSION_DENIED;
 
-	ephy_permission_manager_add (manager, location, EPT_POPUP, permission);
+	ephy_permission_manager_add_permission (manager, location, EPT_POPUP, permission);
 
 	if (allowed)
 	{
@@ -817,6 +831,7 @@ ephy_tab_finalize (GObject *object)
 {
 	EphyTab *tab = EPHY_TAB (object);
 	EphyTabPrivate *priv = tab->priv;
+	guint id = priv->id;
 
 	if (priv->icon != NULL)
 	{
@@ -840,6 +855,14 @@ ephy_tab_finalize (GObject *object)
 #endif
 
 	LOG ("EphyTab finalized %p", tab);
+
+	/* Remove the ID */
+	g_array_index (tabs_id_array, gpointer, id) = NULL;
+	if (--n_tabs == 0)
+	{
+		g_array_free (tabs_id_array, TRUE);
+		tabs_id_array = NULL;
+	}
 }
 
 static gboolean
@@ -1309,47 +1332,63 @@ ephy_tab_file_monitor_cb (GnomeVFSMonitorHandle *handle,
 			  GnomeVFSMonitorEventType event_type,
 			  EphyTab *tab)
 {
+	gboolean uri_is_directory;
+	gboolean should_reload;
+	char* local_path;
 	EphyTabPrivate *priv = tab->priv;
 
 	LOG ("File '%s' has changed, scheduling reload", monitor_uri);
 
+	local_path = gnome_vfs_get_local_path_from_uri(monitor_uri);
+	uri_is_directory = g_file_test(local_path, G_FILE_TEST_IS_DIR);
+	g_free(local_path);
+
 	switch (event_type)
 	{
+		/* These events will always trigger a reload: */
 		case GNOME_VFS_MONITOR_EVENT_CHANGED:
 		case GNOME_VFS_MONITOR_EVENT_CREATED:
-			/* We make a lot of assumptions here, but basically we know
-			 * that we just have to reload, by construction.
-			 * Delay the reload a little bit so we don't endlessly
-			 * reload while a file is written.
-			 */
-			if (priv->reload_delay_ticks == 0)
-			{
-				priv->reload_delay_ticks = 1;
-			}
-			else
-			{
-				/* Exponential backoff */
-				priv->reload_delay_ticks = MIN (priv->reload_delay_ticks * 2,
-								RELOAD_DELAY_MAX_TICKS);
-			}
-
-			if (priv->reload_scheduled_id == 0)
-			{
-				priv->reload_scheduled_id =
-					g_timeout_add (RELOAD_DELAY,
-						       (GSourceFunc) ephy_file_monitor_reload_cb,
-						       tab);
-			}
-
+			should_reload = TRUE;
 			break;
 
+		/* These events will only trigger a reload for directories: */
 		case GNOME_VFS_MONITOR_EVENT_DELETED:
+		case GNOME_VFS_MONITOR_EVENT_METADATA_CHANGED:
+			should_reload = uri_is_directory;
+			break;
+
+		/* These events don't trigger a reload: */
 		case GNOME_VFS_MONITOR_EVENT_STARTEXECUTING:
 		case GNOME_VFS_MONITOR_EVENT_STOPEXECUTING:
-		case GNOME_VFS_MONITOR_EVENT_METADATA_CHANGED:
 		default:
+			should_reload = FALSE;
 			break;
 	}
+
+	if (should_reload) {
+		/* We make a lot of assumptions here, but basically we know
+		 * that we just have to reload, by construction.
+		 * Delay the reload a little bit so we don't endlessly
+		 * reload while a file is written.
+		 */
+		if (priv->reload_delay_ticks == 0)
+		{
+			priv->reload_delay_ticks = 1;
+		}
+		else
+		{
+			/* Exponential backoff */
+			priv->reload_delay_ticks = MIN (priv->reload_delay_ticks * 2,
+					RELOAD_DELAY_MAX_TICKS);
+		}
+
+		if (priv->reload_scheduled_id == 0)
+		{
+			priv->reload_scheduled_id =
+				g_timeout_add (RELOAD_DELAY,
+						(GSourceFunc) ephy_file_monitor_reload_cb, tab);
+		}
+	}
 }
 
 static void
@@ -1359,6 +1398,8 @@ ephy_tab_update_file_monitor (EphyTab *tab,
 	EphyTabPrivate *priv = tab->priv;
 	GnomeVFSMonitorHandle *handle = NULL;
 	gboolean local;
+	char* local_path;
+	GnomeVFSMonitorType monitor_type;
 
 	if (priv->monitor != NULL &&
 	    priv->address != NULL && address != NULL &&
@@ -1374,8 +1415,14 @@ ephy_tab_update_file_monitor (EphyTab *tab,
 	local = g_str_has_prefix (address, "file://");
 	if (local == FALSE) return;
 	
+	local_path = gnome_vfs_get_local_path_from_uri(address);
+	monitor_type = g_file_test(local_path, G_FILE_TEST_IS_DIR)
+		? GNOME_VFS_MONITOR_DIRECTORY
+		: GNOME_VFS_MONITOR_FILE;
+	g_free(local_path);
+
 	if (gnome_vfs_monitor_add (&handle, address,
-				   GNOME_VFS_MONITOR_FILE,
+				   monitor_type,
 				   (GnomeVFSMonitorCallback) ephy_tab_file_monitor_cb,
 				   tab) == GNOME_VFS_OK)
 	{
@@ -1833,28 +1880,26 @@ ephy_tab_visibility_cb (EphyEmbed *embed, gboolean visibility,
 }
 
 static gboolean
-open_link_in_new_tab (EphyTab *tab,
-		      const char *link_address)
+open_link_in_new (EphyTab *tab,
+		  const char *link_address,
+		  guint state)
 {
-	gboolean new_tab;
+	EphyTab *dest;
 
-	new_tab = address_has_web_scheme (link_address);
+	if (!address_has_web_scheme (link_address)) return FALSE;
 
-	if (new_tab)
+	dest = ephy_link_open (EPHY_LINK (tab), link_address, tab,
+			       state & GDK_SHIFT_MASK ? EPHY_LINK_NEW_WINDOW
+						      : EPHY_LINK_NEW_TAB);
+
+	if (dest)
 	{
-		EphyTab *dest;
-		dest = ephy_link_open (EPHY_LINK (tab), link_address,
-				       tab, EPHY_LINK_NEW_TAB);
-
-		if (dest)
-		{
-			ephy_embed_shistory_copy (ephy_tab_get_embed (tab),
-						  ephy_tab_get_embed (dest),
-						  TRUE,   /* back history */
-						  FALSE,  /* forward history */
-					  	  FALSE); /* current index */
-			return TRUE;
-		}
+		ephy_embed_shistory_copy (ephy_tab_get_embed (tab),
+					  ephy_tab_get_embed (dest),
+					  TRUE,   /* back history */
+					  FALSE,  /* forward history */
+					  FALSE); /* current index */
+		return TRUE;
 	}
 
 	return FALSE;
@@ -1870,7 +1915,7 @@ save_property_url (EphyEmbed *embed,
 	const GValue *value;
 	EphyEmbedPersist *persist;
 
-	ephy_embed_event_get_property (event, property, &value);
+	value = ephy_embed_event_get_property (event, property);
 	location = g_value_get_string (value);
 
 	persist = EPHY_EMBED_PERSIST
@@ -1915,7 +1960,8 @@ ephy_tab_dom_mouse_click_cb (EphyEmbed *embed,
 	EphyEmbedEventContext context;
 	guint button, modifier;
 	gboolean handled = TRUE;
-	gboolean with_control, with_shift, is_left_click, is_middle_click;
+	gboolean with_control, with_shift, with_shift_control;
+	gboolean is_left_click, is_middle_click;
 	gboolean is_link, is_image, is_middle_clickable;
 	gboolean middle_click_opens;
 	gboolean is_input;
@@ -1929,8 +1975,9 @@ ephy_tab_dom_mouse_click_cb (EphyEmbed *embed,
 	LOG ("ephy_tab_dom_mouse_click_cb: button %d, context %x, modifier %x",
 	     button, context, modifier);
 
-	with_control = (modifier & GDK_CONTROL_MASK) != 0;
-	with_shift = (modifier & GDK_SHIFT_MASK) != 0;
+	with_control = (modifier & GDK_CONTROL_MASK) == GDK_CONTROL_MASK;
+	with_shift = (modifier & GDK_SHIFT_MASK) == GDK_SHIFT_MASK;
+	with_shift_control = (modifier & (GDK_SHIFT_MASK | GDK_CONTROL_MASK)) == (GDK_SHIFT_MASK | GDK_CONTROL_MASK);
 	is_left_click = (button == 1);
 	is_middle_click = (button == 2);
 
@@ -1946,14 +1993,16 @@ ephy_tab_dom_mouse_click_cb (EphyEmbed *embed,
 	is_input = (context & EPHY_EMBED_CONTEXT_INPUT) != 0;
 
 	/* ctrl+click or middle click opens the link in new tab */
-	if (is_link && ((is_left_click && with_control) || is_middle_click))
+	if (is_link &&
+	    ((is_left_click && (with_control || with_shift_control)) ||
+	     is_middle_click))
 	{
 		const GValue *value;
 		const char *link_address;
 
-		ephy_embed_event_get_property (event, "link", &value);
+		value = ephy_embed_event_get_property (event, "link");
 		link_address = g_value_get_string (value);
-		handled = open_link_in_new_tab (tab, link_address);
+		handled = open_link_in_new (tab, link_address, modifier);
 	}
 	/* shift+click saves the link target */
 	else if (is_link && is_left_click && with_shift)
@@ -2009,11 +2058,41 @@ ephy_tab_init (EphyTab *tab)
 	EphyTabPrivate *priv;
 	GObject *embed;
 	EphyFaviconCache *cache;
+	guint id;
 
 	LOG ("EphyTab initialising %p", tab);
 
 	priv = tab->priv = EPHY_TAB_GET_PRIVATE (tab);
 
+	/* Make tab ID */
+	++n_tabs;
+
+	if (tabs_id_array == NULL)
+	{
+		tabs_id_array = g_array_sized_new (FALSE /* zero-terminate */,
+					       TRUE /* clear */,
+					       sizeof (gpointer),
+					       64 /* initial size */);
+	}
+
+	for (id = 0; id < tabs_id_array->len; ++id)
+	{
+		if (g_array_index (tabs_id_array, gpointer, id) == NULL) break;
+	}
+
+	priv->id = id;
+
+	/* Grow array if necessary */
+	if (id >= tabs_id_array->len)
+	{
+		g_array_append_val (tabs_id_array, tab);
+		g_assert (g_array_index (tabs_id_array, gpointer, id) == tab);
+	}
+	else
+	{
+		g_array_index (tabs_id_array, gpointer, id) = tab;
+	}
+
 	tab->priv->total_requests = 0;
 	tab->priv->cur_requests = 0;
 	tab->priv->width = -1;
@@ -2436,3 +2515,12 @@ ephy_tab_get_zoom (EphyTab *tab)
 
 	return tab->priv->zoom;
 }
+
+/* private */
+guint
+_ephy_tab_get_id (EphyTab *tab)
+{
+	g_return_val_if_fail (EPHY_IS_TAB (tab), (guint) -1);
+
+	return tab->priv->id;
+}
